from ....engines.daft import Daft
from ....engines.delta_rs import DeltaRs
import posixpath

class DaftELTBench:
    def __init__(self, engine : Daft):
        self.engine = engine

        self.delta_rs = DeltaRs()
        self.write_deltalake = self.delta_rs.write_deltalake
        self.DeltaTable = self.delta_rs.DeltaTable

    def create_total_sales_fact(self):
        fact_table_df = (
            self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'store_sales'))
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'date_dim')), 
                left_on="ss_sold_date_sk", 
                right_on="d_date_sk"
            )
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'store')), 
                left_on="ss_store_sk", 
                right_on="s_store_sk"
            )
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'item')), 
                left_on="ss_item_sk", 
                right_on="i_item_sk"
            )
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'customer')), 
                left_on="ss_customer_sk", 
                right_on="c_customer_sk"
            )
            .with_columns({
                "sale_date": self.engine.daft.col("d_date")
            })
            .where(self.engine.daft.col("d_year") == 2001)
            .groupby(["s_store_id", "i_item_id", "c_customer_id", "sale_date"])
            .agg([
                self.engine.daft.col("ss_quantity").sum().alias("total_quantity"),
                self.engine.daft.col("ss_net_paid").sum().cast(self.engine.daft.DataType.decimal128(38, 2)).alias("total_net_paid"),
                self.engine.daft.col("ss_net_profit").sum().cast(self.engine.daft.DataType.decimal128(38, 2)).alias("total_net_profit")
            ])
            .sort(["s_store_id", "sale_date"])
        )

        fact_table_df.write_deltalake(
                posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact'),
                mode="overwrite"
            ) 


    def merge_percent_into_total_sales_fact(self, percent: float):
        
        sampled_fact_data = (
            self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'store_sales'))
            .sample(percent)
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'date_dim')), 
                left_on="ss_sold_date_sk", 
                right_on="d_date_sk"
            )
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'store')), 
                left_on="ss_store_sk", 
                right_on="s_store_sk"
            )
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'item')), 
                left_on="ss_item_sk", 
                right_on="i_item_sk"
            )
            .join(
                self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'customer')), 
                left_on="ss_customer_sk", 
                right_on="c_customer_sk"
            )
            .with_columns({
                # Use hash for pseudo-random values
                "hash_val": (self.engine.daft.col("ss_customer_sk") + self.engine.daft.col("ss_sold_date_sk")).hash(),
                "s_store_id": self.engine.daft.col("s_store_id"),
                "i_item_id": self.engine.daft.col("i_item_id"),
                "sale_date": self.engine.daft.col("d_date"),
            })
            .with_columns({
                "c_customer_id": (
                    (self.engine.daft.col("hash_val") % 2 == 0)
                    .if_else(
                        self.engine.daft.col("c_customer_id"),
                        (self.engine.daft.col("hash_val") % 10000).cast(self.engine.daft.DataType.string())
                    )
                ),
                "total_quantity": self.engine.daft.col("ss_quantity") + (self.engine.daft.col("hash_val") % 5 + 1),
                "total_net_paid": (self.engine.daft.col("ss_net_paid") + ((self.engine.daft.col("hash_val") % 5000) / 100.0 + 5)).cast(self.engine.daft.DataType.decimal128(38, 2)),
                "total_net_profit": (self.engine.daft.col("ss_net_profit") + ((self.engine.daft.col("hash_val") % 2000) / 100.0 + 1)).cast(self.engine.daft.DataType.decimal128(38, 2))
            })
            .select(
                "s_store_id",
                "i_item_id", 
                "c_customer_id",
                "sale_date",
                "total_quantity",
                "total_net_paid",
                "total_net_profit"
            )
            .to_arrow()
        )

        fact_table = self.DeltaTable(
            posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact')
        )
    
        fact_table.merge(
            source=sampled_fact_data,
            predicate="""
            target.s_store_id = source.s_store_id AND 
            target.i_item_id = source.i_item_id AND 
            target.c_customer_id = source.c_customer_id AND 
            target.sale_date = source.sale_date
            """,
            source_alias="source",
            target_alias="target"
        ) \
        .when_matched_update({
            "total_quantity": "target.total_quantity + source.total_quantity",
            "total_net_paid": "target.total_net_paid + source.total_net_paid",
            "total_net_profit": "target.total_net_profit + source.total_net_profit",
        }) \
        .when_not_matched_insert({
            "s_store_id": "source.s_store_id",
            "i_item_id": "source.i_item_id",
            "c_customer_id": "source.c_customer_id",
            "sale_date": "source.sale_date",
            "total_quantity": "source.total_quantity",
            "total_net_paid": "source.total_net_paid",
            "total_net_profit": "source.total_net_profit",
        }).execute()

    def query_total_sales_fact(self):
        query_df = (
            self.engine.daft.read_deltalake(posixpath.join(self.engine.delta_mount_schema_path, 'total_sales_fact'))
                .groupby(self.engine.daft.col("sale_date").dt.year())
                .agg(self.engine.daft.col("total_net_profit").sum().alias("sum_net_profit"))
                .to_pandas()
        )


import posixpath
import os
import importlib.util


class _TPCDataGenerator:
    """
    Base class for TPC data generation. PLEASE DO NOT INSTANTIATE THIS CLASS DIRECTLY. Use the TPCHDataGenerator and TPCDSDataGenerator
    subclasses instead.
    """
    GEN_UTIL = ''

    def __init__(self, scale_factor: int, target_mount_folder_path: str = None, target_row_group_size_mb: int = 128):
        """
        Initialize the TPC data generator with a scale factor.

        :param scale_factor: The scale factor for the data generation.
        """
        self.scale_factor = scale_factor
        self.target_mount_folder_path = target_mount_folder_path
        self.target_row_group_size_mb = target_row_group_size_mb

        if importlib.util.find_spec("duckdb") is None:
            raise ImportError(
                "DuckDB is used for data generation but is not installed. Install using `%pip install lakebench[duckdb]` or `%pip install lakebench[datagen]`"
            )

    def run(self):
        """
        This method uses DuckDB to generate in-memory tables based on the specified 
        scale factor and writes them to Parquet files. It estimates the average row 
        size in MB using a sample of the data since DuckDB only supports specifying 
        the number of rows per row group. The generated tables are written to the 
        specified target folder with optimized row group sizes.

        Parameters
        ----------
        None
        
        Notes
        -----
        - The method creates a sample Parquet file for each table to estimate row sizes.
        - The full table is then written as Parquet files with optimized row group sizes.
        - Temporary files and in-memory tables are cleaned up after processing.
        """
        import duckdb
        import pyarrow.parquet as pq

        con = duckdb.connect()
        print(f"Generating in-memory tables")
        con.execute(f"CALL {self.GEN_UTIL}(sf={self.scale_factor})")
        tables = [row[0] for row in con.execute("SHOW TABLES").fetchall()]
        print(f"Generated in-memory tables: {tables}")

        for table in tables:
            sample_file = posixpath.join(self.target_mount_folder_path, f"{table}_sample.parquet")
            full_folder_path = posixpath.join(self.target_mount_folder_path, table)
            os.makedirs(self.target_mount_folder_path, exist_ok=True)
            # Write a sample for row size estimation
            print(f"\nSampling {table} to evaluate row count to target {self.target_row_group_size_mb}mb row groups...")
            con.execute(f"""
                COPY (SELECT * FROM {table} LIMIT 1000000)
                TO '{sample_file}'
                (FORMAT 'parquet')
            """)

            pf = pq.ParquetFile(sample_file)
            rg = pf.metadata.row_group(0)
            avg_row_size = rg.total_byte_size / rg.num_rows
            #print(f"{table} sample: {rg.num_rows} rows, {rg.total_byte_size / (1024*1024):.2f} MB")
            #print(f"Avg row size: {avg_row_size:.2f} bytes")
            target_size_bytes = self.target_row_group_size_mb * 1024 * 1024
            target_rows = int(target_size_bytes / avg_row_size)
            #print(f"Target ROW_GROUP_SIZE for ~{self.target_row_group_size_mb} MB: {target_rows} rows")

            # Write full table
            print(f"Writing {table} to {full_folder_path} with ROW_GROUP_SIZE {target_rows}...")
            con.execute(f"""
                COPY {table} TO '{full_folder_path}'
                (FORMAT 'parquet', ROW_GROUP_SIZE {target_rows}, PER_THREAD_OUTPUT)
            """)

            con.execute(f"DROP TABLE {table}")
            os.remove(sample_file)


from .base import BaseEngine
from  .delta_rs import DeltaRs

import posixpath

class DuckDB(BaseEngine):
    """
    DuckDB Engine for ELT Benchmarks.
    """
    SQLGLOT_DIALECT = "duckdb"
    REQUIRED_READ_ENDPOINT = None
    REQUIRED_WRITE_ENDPOINT = "abfss"

    def __init__(
            self, 
            delta_abfss_schema_path: str
            ):
        """
        Initialize the DuckDB Engine Configs
        """
        import duckdb
        self.duckdb = duckdb
        self.delta_abfss_schema_path = delta_abfss_schema_path
        self.deltars = DeltaRs()
        self.catalog_name = None
        self.schema_name = None

    def load_parquet_to_delta(self, parquet_folder_path: str, table_name: str):
        arrow_df = self.duckdb.sql(f""" FROM parquet_scan('{posixpath.join(parquet_folder_path, '*.parquet')}') """).record_batch()
        self.deltars.write_deltalake(
            posixpath.join(self.delta_abfss_schema_path, table_name),
            arrow_df,
            mode="overwrite",
            engine='pyarrow'
        )  

    def register_table(self, table_name: str):
        """
        Register a Delta table in DuckDB.
        """
        self.duckdb.sql(f"""
            CREATE OR REPLACE VIEW {table_name} 
            AS SELECT * FROM delta_scan('{posixpath.join(self.delta_abfss_schema_path, table_name)}')
        """)

    def execute_sql_query(self, query: str):
        """
        Execute a SQL query using DuckDB.
        """
        result = self.duckdb.sql(query).df()

    def optimize_table(self, table_name: str):
        fact_table = self.deltars.DeltaTable(
            posixpath.join(self.delta_abfss_schema_path, table_name)
        )
        fact_table.optimize.compact()

    def vacuum_table(self, table_name: str, retain_hours: int = 168, retention_check: bool = True):
        fact_table = self.deltars.DeltaTable(
            posixpath.join(self.delta_abfss_schema_path, table_name)
        )
        fact_table.vacuum(retain_hours, enforce_retention_duration=retention_check, dry_run=False)


from typing import Optional
from ..base import BaseBenchmark

from .engine_impl.spark import SparkELTBench
from .engine_impl.duckdb import DuckDBELTBench
from .engine_impl.daft import DaftELTBench
from .engine_impl.polars import PolarsELTBench

from ...engines.base import BaseEngine
from ...engines.spark import Spark
from ...engines.duckdb import DuckDB
from ...engines.daft import Daft
from ...engines.polars import Polars

import posixpath


class ELTBench(BaseBenchmark):
    """
    Class for running the ELTBench benchmark.

    The ELTBench benchmark is designed to evaluate end-to-end performance of engines in supporting typical Lakehouse architecture patterns. This includes bulk loading data, creation of star schema tables, incrementally merging data into tables, performing maintenance jobs, and running ad-hoc aggregation queries. Supported engines are listed in the `self.BENCHMARK_IMPL_REGISTRY` constant. ELTBench supports two modes: 'light' and 'full'. The 'light' mode represents a small workload, while the 'full' mode includes a larger scope of tests.

    Parameters
    ----------
    engine : BaseEngine
        The engine to use for executing the benchmark.
    scenario_name : str
        The name of the benchmark scenario.
    tpcds_parquet_mount_path : str, optional
        Path to the mounted TPC-DS parquet files. Must be the root directory containing a folder named after each table in TABLE_REGISTRY. If not provided, `tpcds_parquet_abfss_path` must be specified assuming the engine supports ABFSS.
    tpcds_parquet_abfss_path : str, optional
        Path to the parquet files in ABFSS. Must be the root directory containing a folder named after each table in TABLE_REGISTRY.
    result_abfss_path : str, optional
        ABFSS path to the table where results will be saved. Must be specified if `save_results` is True.
    save_results : bool, optional
        Whether to save the benchmark results. Results can also be accessed via the `self.results` attribute after running the benchmark.

    Methods
    -------
    run(mode='light')
        Runs the benchmark in the specified mode. Valid modes are 'light' and 'full'.
    run_light_mode()
        Executes the 'light' mode of the benchmark, including data loading, table creation, incremental merging, maintenance jobs, and ad-hoc queries.
    """

    BENCHMARK_IMPL_REGISTRY = {
        Spark: SparkELTBench,
        DuckDB: DuckDBELTBench,
        Daft: DaftELTBench,
        Polars: PolarsELTBench
    }
    MODE_REGISTRY = ['light', 'full']
    TABLE_REGISTRY = [
        'call_center', 'catalog_page', 'catalog_returns', 'catalog_sales',
        'customer', 'customer_address', 'customer_demographics', 'date_dim',
        'household_demographics', 'income_band', 'inventory', 'item',
        'promotion', 'reason', 'ship_mode', 'store', 'store_returns',
        'store_sales', 'time_dim', 'warehouse', 'web_page', 'web_returns',
        'web_sales', 'web_site'
    ]

    def __init__(
            self, 
            engine: BaseEngine, 
            scenario_name: str,
            tpcds_parquet_mount_path: Optional[str] = None,
            tpcds_parquet_abfss_path: Optional[str] = None,
            result_abfss_path: Optional[str] = None,
            save_results: bool = False
            ):
        super().__init__(engine, scenario_name, result_abfss_path, save_results)
        for base_engine, benchmark_impl in self.BENCHMARK_IMPL_REGISTRY.items():
            if isinstance(engine, base_engine):
                self.benchmark_impl_class = benchmark_impl
                if self.benchmark_impl_class is None:
                    raise ValueError(
                        f"No benchmark implementation registered for engine type: {type(engine).__name__} "
                        f"in benchmark '{self.__class__.__name__}'."
                    )
                break
        else:
            raise ValueError(
                f"No benchmark implementation registered for engine type: {type(engine).__name__} "
                f"in benchmark '{self.__class__.__name__}'."
            )
        
        if isinstance(engine, Daft):
            if tpcds_parquet_mount_path is None:
                raise ValueError("parquet_mount_path must be provided for Daft engine.")
        self.source_data_path = tpcds_parquet_mount_path or tpcds_parquet_abfss_path
        self.engine = engine
        self.scenario_name = scenario_name
        self.benchmark_impl = self.benchmark_impl_class(
            self.engine
        )

        match engine.REQUIRED_READ_ENDPOINT:
            case 'mount':
                if tpcds_parquet_mount_path is None:
                    raise ValueError(f"parquet_mount_path must be provided for {type(engine).__name__} engine.")
                self.source_data_path = tpcds_parquet_mount_path
            case 'abfss':
                if tpcds_parquet_abfss_path is None:
                    raise ValueError(f"parquet_abfss_path must be provided for {type(engine).__name__} engine.")
                self.source_data_path = tpcds_parquet_abfss_path
            case _:
                if tpcds_parquet_mount_path is None and tpcds_parquet_abfss_path is None:
                    raise ValueError(
                        f"Either parquet_mount_path or parquet_abfss_path must be provided for {type(engine).__name__} engine."
                    )
                self.source_data_path = tpcds_parquet_abfss_path or tpcds_parquet_mount_path

    def run(self, mode: str = 'light'):
        """
        Executes the benchmark in the specified mode.
        
        Parameters
        ----------
        mode : str, optional
            The mode in which to run the benchmark. Supported modes are:
            - 'light': Runs the benchmark in light mode.
            - 'full': Placeholder for full mode, which is not implemented yet.
        """

        match mode:
            case 'light':
                self.run_light_mode()
            case 'full':
                raise NotImplementedError("Full mode is not implemented yet.")
            case _:
                raise ValueError(f"Mode '{mode}' is not supported. Supported modes: {self.MODE_REGISTRY}.")

    def run_light_mode(self):
        """
        Executes the light mode benchmark workflow for processing and querying data.
        This method performs a series of operations on data tables, including loading data 
        from parquet files into Delta tables, creating a fact table, merging data, optimizing 
        the table, vacuuming the table, and running an ad-hoc query. The results are posted 
        at the end of the workflow.

        Parameters
        ----------
        None
        """
        for table_name in ('store_sales', 'date_dim', 'store', 'item', 'customer'):
            with self.timer(phase="Read parquet, write delta (x5)", test_item=table_name, engine=self.engine):
                self.engine.load_parquet_to_delta(
                    parquet_folder_path=posixpath.join(self.source_data_path, table_name), 
                    table_name=table_name
                )
        with self.timer(phase="Create fact table", test_item='total_sales_fact', engine=self.engine):
            self.benchmark_impl.create_total_sales_fact()

        for _ in range(3):
            with self.timer(phase="Merge 0.1% into fact table (3x)", test_item='total_sales_fact', engine=self.engine):
                self.benchmark_impl.merge_percent_into_total_sales_fact(0.001)

        with self.timer(phase="OPTIMIZE", test_item='total_sales_fact', engine=self.engine):
            self.engine.optimize_table('total_sales_fact')

        with self.timer(phase="VACUUM", test_item='total_sales_fact', engine=self.engine):
            self.engine.vacuum_table('total_sales_fact', retain_hours=0, retention_check=False)

        with self.timer(phase="Ad-hoc query (small result aggregation)", test_item='total_sales_fact', engine=self.engine):
            self.benchmark_impl.query_total_sales_fact()

        self.post_results()




from ....engines.spark import Spark

class SparkELTBench:
    def __init__(self, engine: Spark):
        
        self.engine = engine
        self.engine.create_schema_if_not_exists(drop_before_create=True)

    def create_total_sales_fact(self):
        self.engine.spark.sql("""
            CREATE OR REPLACE TABLE total_sales_fact AS
            SELECT 
                s.s_store_id,
                i.i_item_id,
                c.c_customer_id,
                d.d_date AS sale_date,
                SUM(ss.ss_quantity) AS total_quantity,
                SUM(ss.ss_net_paid) AS total_net_paid,
                SUM(ss.ss_net_profit) AS total_net_profit
            FROM 
                store_sales ss
            JOIN 
                date_dim d ON ss.ss_sold_date_sk = d.d_date_sk
            JOIN 
                store s ON ss.ss_store_sk = s.s_store_sk
            JOIN 
                item i ON ss.ss_item_sk = i.i_item_sk
            JOIN 
                customer c ON ss.ss_customer_sk = c.c_customer_sk
            WHERE 
                d.d_year = 2001
            GROUP BY 
                s.s_store_id, i.i_item_id, c.c_customer_id, d.d_date
            ORDER BY 
                s.s_store_id, d.d_date;
        """)

    def merge_percent_into_total_sales_fact(self, percent: float):
        store_sales = self.engine.spark.table("store_sales")
        date_dim = self.engine.spark.table("date_dim")
        store = self.engine.spark.table("store")
        item = self.engine.spark.table("item")
        customer = self.engine.spark.table("customer")

        sampled_data = store_sales.sample(fraction=percent)

        sampled_fact_data = (
            sampled_data.join(date_dim, store_sales.ss_sold_date_sk == date_dim.d_date_sk)
            .join(store, store_sales.ss_store_sk == store.s_store_sk)
            .join(item, store_sales.ss_item_sk == item.i_item_sk)
            .join(customer, store_sales.ss_customer_sk == customer.c_customer_sk)
            .select(
                store.s_store_id,
                item.i_item_id,
                customer.c_customer_id,
                date_dim.d_date.alias("sale_date"),
                store_sales.ss_quantity.alias("total_quantity"),
                store_sales.ss_net_paid.alias("total_net_paid"),
                store_sales.ss_net_profit.alias("total_net_profit")
            )
        )

        synthetic_merge_data = sampled_fact_data \
            .withColumn("c_customer_id", self.engine.sf.when(self.engine.sf.rand() > 0.5, self.engine.sf.expr("CAST(rand() * 100000 AS INT)")).otherwise(sampled_fact_data.c_customer_id)) \
            .withColumn("total_quantity", sampled_fact_data.total_quantity + self.engine.sf.expr("floor(rand() * 5 + 1)")) \
            .withColumn("total_net_paid", sampled_fact_data.total_net_paid + self.engine.sf.expr("rand() * 50 + 5")) \
            .withColumn("total_net_profit", sampled_fact_data.total_net_profit + self.engine.sf.expr("rand() * 20 + 1"))

        # Register as a temporary view for SQL-based merge
        synthetic_merge_data.createOrReplaceTempView("synthetic_merge_data")

        self.engine.spark.sql("""
            MERGE INTO total_sales_fact AS target
            USING (
                SELECT 
                    s_store_id, 
                    i_item_id, 
                    c_customer_id, 
                    sale_date, 
                    total_quantity, 
                    total_net_paid, 
                    total_net_profit
                FROM synthetic_merge_data
            ) AS source
            ON 
                target.s_store_id = source.s_store_id AND 
                target.i_item_id = source.i_item_id AND 
                target.c_customer_id = source.c_customer_id AND 
                target.sale_date = source.sale_date
            WHEN MATCHED THEN 
                UPDATE SET 
                    target.total_quantity = target.total_quantity + source.total_quantity,
                    target.total_net_paid = target.total_net_paid + source.total_net_paid,
                    target.total_net_profit = target.total_net_profit + source.total_net_profit
            WHEN NOT MATCHED THEN 
                INSERT (s_store_id, i_item_id, c_customer_id, sale_date, total_quantity, total_net_paid, total_net_profit)
                VALUES (source.s_store_id, source.i_item_id, source.c_customer_id, source.sale_date, source.total_quantity, source.total_net_paid, source.total_net_profit);
        """)
        
    def query_total_sales_fact(self):
        df = self.engine.spark.sql(f"""
                            select sum(total_net_profit), year(sale_date) 
                            from total_sales_fact group by year(sale_date)
                            """)
        result = df.collect()


import time
from datetime import datetime
from contextlib import contextmanager
from ..engines.spark import Spark

@contextmanager
def timer(phase: str = "Elapsed time", test_item: str = '', engine: str = None):
    if not hasattr(timer, "results"):
        timer.results = []

    iteration = sum(1 for result in timer.results if result[0] == phase and result[1] == test_item) + 1

    if isinstance(engine, Spark):
        engine.spark.sparkContext.setJobDescription(f"{phase} - {test_item} [i:{iteration}]")

    start = time.time()
    start_datetime = datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
    success = True
    error_message = None
    error_type = None

    try:
        yield
    except Exception as e:
        success = False
        error_message = str(e)
        error_type = type(e).__name__  # Capture the error type
        print(f"Error during {phase} - {test_item}... {error_type}: {error_message}")
        
    finally:
        end = time.time()
        duration = int((end - start) * 1000)
        print(f"{phase} - {test_item}{f' [i:{iteration}]' if iteration > 1 else ''}: {(duration / 1000):.2f} seconds")
        if isinstance(engine, Spark):
            engine.spark.sparkContext.setJobDescription(None)
        timer.results.append((phase, test_item, start_datetime, duration, iteration, success, f"{error_type}: {error_message}" if error_message else ''))

def _clear_results():
    if hasattr(timer, "results"):
        timer.results = []

timer.clear_results = _clear_results


from ....engines.duckdb import DuckDB
from ....engines.delta_rs import DeltaRs

import posixpath

class DuckDBELTBench:
    def __init__(self, engine : DuckDB):
        self.engine = engine

        self.delta_rs = DeltaRs()
        self.write_deltalake = self.delta_rs.write_deltalake
        self.DeltaTable = self.delta_rs.DeltaTable

    def create_total_sales_fact(self):
        self.engine.duckdb.sql("use main")

        for table in ['store_sales', 'date_dim', 'store', 'item', 'customer']:
            self.engine.register_table(table)

        arrow_df = self.engine.duckdb.sql("""
        SELECT 
            s.s_store_id,
            i.i_item_id,
            c.c_customer_id,
            d.d_date AS sale_date,
            SUM(ss.ss_quantity) AS total_quantity,
            SUM(ss.ss_net_paid) AS total_net_paid,
            SUM(ss.ss_net_profit) AS total_net_profit
        FROM 
            store_sales ss
        JOIN 
            date_dim d ON ss.ss_sold_date_sk = d.d_date_sk
        JOIN 
            store s ON ss.ss_store_sk = s.s_store_sk
        JOIN 
            item i ON ss.ss_item_sk = i.i_item_sk
        JOIN 
            customer c ON ss.ss_customer_sk = c.c_customer_sk
        WHERE 
            d.d_year = 2001
        GROUP BY 
            s.s_store_id, i.i_item_id, c.c_customer_id, d.d_date
        ORDER BY 
            s.s_store_id, d.d_date;

        """).record_batch()

        self.write_deltalake(
            posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact'),
            arrow_df,
            mode="overwrite",
            engine='pyarrow'
        )

    def merge_percent_into_total_sales_fact(self, percent: float):
        self.engine.duckdb.sql("use main")

        for table in ['store_sales', 'date_dim', 'store', 'item', 'customer']:
            self.engine.register_table(table)

        percent_str = f"{percent * 100:.1f}%"

        synthetic_data = self.engine.duckdb.sql(f"""
            SELECT 
                s_store_id, 
                i_item_id, 
                case when random() > 0.5 then c_customer_id else CAST(10000 * random() AS string) end AS c_customer_id,
                d.d_date as sale_date,
                ss_quantity + floor(random() * 5 + 1) AS total_quantity,
                ss_net_paid + random() * 50 + 5 AS total_net_paid,
                ss_net_profit + random() * 20 + 1 AS total_net_profit
            FROM 
                (SELECT * FROM store_sales USING SAMPLE {percent_str}) ss
            JOIN 
                date_dim d ON ss.ss_sold_date_sk = d.d_date_sk
            JOIN 
                store s ON ss.ss_store_sk = s.s_store_sk
            JOIN 
                item i ON ss.ss_item_sk = i.i_item_sk
            JOIN 
                customer c ON ss.ss_customer_sk = c.c_customer_sk

        """).record_batch()

        fact_table = self.DeltaTable(posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact'))

        fact_table.merge(
                source=synthetic_data,
                predicate="""
                target.s_store_id = source.s_store_id AND 
                target.i_item_id = source.i_item_id AND 
                target.c_customer_id = source.c_customer_id AND 
                target.sale_date = source.sale_date
                """,
                source_alias="source",
                target_alias="target"
            ) \
            .when_matched_update(
                {
                    "total_quantity": "target.total_quantity + source.total_quantity",
                    "total_net_paid": "target.total_net_paid + source.total_net_paid",
                    "total_net_profit": "target.total_net_profit + source.total_net_profit",
                }
            ) \
            .when_not_matched_insert(
                {
                    "s_store_id": "source.s_store_id",
                    "i_item_id": "source.i_item_id",
                    "c_customer_id": "source.c_customer_id",
                    "sale_date": "source.sale_date",
                    "total_quantity": "source.total_quantity",
                    "total_net_paid": "source.total_net_paid",
                    "total_net_profit": "source.total_net_profit",
                }
            ) \
            .execute()

    def query_total_sales_fact(self):
        df = self.engine.duckdb.sql(f"""
            select sum(total_net_profit), year(sale_date) 
            from delta_scan('{posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact')}') group by year(sale_date)
        """).df()


from .base import BaseEngine
import os

class Spark(BaseEngine):
    """
    Spark Engine for ELT Benchmarks.
    """
    SQLGLOT_DIALECT = "spark"
    REQUIRED_READ_ENDPOINT = None
    REQUIRED_WRITE_ENDPOINT = None

    def __init__(
            self,
            catalog_name: str,
            schema_name: str,
            spark_measure_telemetry: bool = False
            ):
        """
        Initialize the SparkEngine with a Spark session.
        """
        from pyspark.sql import SparkSession
        if spark_measure_telemetry:
            from sparkmeasure import StageMetrics
        self.spark_measure_telemetry = spark_measure_telemetry

        import pyspark.sql.functions as sf
        self.sf = sf
        self.spark = SparkSession.builder.getOrCreate()

        self.catalog_name = catalog_name
        self.schema_name = schema_name
        self.full_catalog_schema_reference : str = f"`{self.catalog_name}`.`{self.schema_name}`"

    def create_schema_if_not_exists(self, drop_before_create: bool = True):
        """
        Prepare an empty schema in the lakehouse.
        """
        if drop_before_create:
            self.spark.sql(f"DROP SCHEMA IF EXISTS {self.full_catalog_schema_reference}")
        self.spark.sql(f"CREATE SCHEMA IF NOT EXISTS {self.full_catalog_schema_reference}")
        self.spark.sql(f"USE {self.full_catalog_schema_reference}")

    def append_array_to_delta(self, abfss_path: str, array: list):
        """
        Append an array to a Delta table.
        """
        # Use default order of columns in dictionary
        columns = list(array[0].keys())
        df = self.spark.createDataFrame(array).select(*columns)
        df.write.option("mergeSchema", "true").option("delta.enableDeletionVectors", "false").format("delta").mode("append").save(abfss_path)

    def get_total_cores(self) -> int:
        """
        Returns the total number of CPU cores available in the Spark cluster.
        
        Assumes that the driver and workers nodes are all the same VM size.
        """
        cores = int(len(set(executor.host() for executor in self.spark.sparkContext._jsc.sc().statusTracker().getExecutorInfos())) * os.cpu_count())
        return cores
    
    def get_compute_size(self) -> str:
        """
        Returns a formatted string with the compute size.
        
        Assumes that the driver and workers nodes are all the same VM size.
        """        
        sc_conf_dict = {key: value for key, value in self.spark.sparkContext.getConf().getAll()}
        executor_count = self.spark.sparkContext._jsc.sc().getExecutorMemoryStatus().size() - 1
        executor_cores = int(sc_conf_dict.get('spark.executor.cores'))
        vm_host_count = len(set(executor.host() for executor in self.spark.sparkContext._jsc.sc().statusTracker().getExecutorInfos()))
        worker_count = vm_host_count - 1
        worker_cores = os.cpu_count()
        as_min_workers = sc_conf_dict['spark.dynamicAllocation.initialExecutors'] if sc_conf_dict.get('spark.autoscale.executorResourceInfoTag.enabled', 'false') == 'true' else None
        as_max_workers = sc_conf_dict['spark.dynamicAllocation.maxExecutors'] if sc_conf_dict.get('spark.autoscale.executorResourceInfoTag.enabled', 'false') == 'true' else None
        as_enabled = True if as_min_workers != as_max_workers and sc_conf_dict.get('spark.dynamicAllocation.minExecutors', None) != sc_conf_dict.get('spark.dynamicAllocation.maxExecutors', None) else False
        type = "SingleNode" if vm_host_count == 1 and not as_enabled else 'MultiNode'
        workers_word = 'Workers' if worker_count > 1 or int(as_max_workers) > 1  else 'Worker'
        executors_per_worker = int(executor_count / worker_count) if worker_count > 0 else 1
        executors_word = 'Executors' if executors_per_worker > 1 else 'Executor'
        executor_str = f"({executors_per_worker} x {executor_cores}vCore {executors_word}{' ea.' if type != 'SingleNode' else ''})"

        if type == 'SingleNode':
            cluster_config = f"{worker_cores}vCore {type} {executor_str}"
        elif as_enabled:
            cluster_config = f"{as_min_workers}-{as_max_workers} x {worker_cores}vCore {workers_word} {executor_str}"
        else:
            cluster_config = f"{worker_count} x {worker_cores}vCore {workers_word} {executor_str}"

        return cluster_config
    
    def load_parquet_to_delta(self, parquet_folder_path: str, table_name: str):
        df = self.spark.read.parquet(parquet_folder_path)
        df.write.mode("append").saveAsTable(table_name)
    
    def execute_sql_query(self, query: str):
        execute_sql = self.spark.sql(query).collect()
    
    def execute_sql_statement(self, statement: str):
        """
        Execute a SQL statement.
        """
        self.spark.sql(statement)

    def optimize_table(self, table_name: str):
        self.spark.sql(f"OPTIMIZE {self.full_catalog_schema_reference}.{table_name}")

    def vacuum_table(self, table_name: str, retain_hours: int = 168, retention_check: bool = True):
        self.spark.conf.set("spark.databricks.delta.retentionDurationCheck.enabled", retention_check)
        self.spark.sql(f"VACUUM {self.full_catalog_schema_reference}.{table_name} RETAIN {retain_hours} HOURS")



from .base import BaseEngine
from .delta_rs import DeltaRs

from IPython.core.getipython import get_ipython
notebookutils = get_ipython().user_ns.get("notebookutils")

import posixpath

class Polars(BaseEngine):
    """
    Polars Engine for ELT Benchmarks.
    """
    SQLGLOT_DIALECT = "duckdb"
    REQUIRED_READ_ENDPOINT = None
    REQUIRED_WRITE_ENDPOINT = "abfss"

    def __init__(
            self, 
            delta_abfss_schema_path: str
            ):
        """
        Initialize the Polars Engine Configs
        """
        import polars as pl
        self.pl = pl
        self.delta_abfss_schema_path = delta_abfss_schema_path
        self.deltars = DeltaRs()
        self.storage_options={
            "bearer_token": notebookutils.credentials.getToken('storage')
        }
        self.catalog_name = None
        self.schema_name = None
        self.sql = pl.SQLContext()

    def load_parquet_to_delta(self, parquet_folder_path: str, table_name: str):
        table_df = self.pl.scan_parquet(
            posixpath.join(parquet_folder_path, '*.parquet'), 
            storage_options=self.storage_options
        )
        table_df.collect(engine='streaming').write_delta(
            posixpath.join(self.delta_abfss_schema_path, table_name), 
            mode="overwrite", 
            storage_options=self.storage_options
        )

    def register_table(self, table_name: str):
        """
        Register a Delta table LazyFrame in Polars.
        """
        df = self.pl.scan_delta(
            posixpath.join(self.delta_abfss_schema_path, table_name), 
            storage_options=self.storage_options
        )
        self.sql.register(table_name, df)

    def execute_sql_query(self, query: str):
        """
        Execute a SQL query using Polars.
        """
        result = self.sql.execute(query).collect(engine='streaming')

    def optimize_table(self, table_name: str):
        fact_table = self.deltars.DeltaTable(
            posixpath.join(self.delta_abfss_schema_path, table_name)
        )
        fact_table.optimize.compact()

    def vacuum_table(self, table_name: str, retain_hours: int = 168, retention_check: bool = True):
        fact_table = self.deltars.DeltaTable(
            posixpath.join(self.delta_abfss_schema_path, table_name)
        )
        fact_table.vacuum(retain_hours, enforce_retention_duration=retention_check, dry_run=False)


from .._tpc import _TPC

class TPCDS(_TPC):
    """
    Class for running the TPC-DS benchmark.

    This class provides functionality for running the TPC-DS benchmark, including loading data,
    executing queries, and performing power tests. Supported engines are listed in the 
    `self.BENCHMARK_IMPL_REGISTRY` constant.

    Parameters
    ----------
    engine : object
        The engine to use for executing the benchmark.
    scenario_name : str
        The name of the benchmark scenario.
    query_list : list of str, optional
        List of queries to execute. Use '*' for all queries. If not specified, all queries will be run.
    parquet_mount_path : str, optional
        Path to the mounted parquet files. Must be the root directory containing a folder named after 
        each table in TABLE_REGISTRY. If not provided, `parquet_abfss_path` must be specified assuming 
        the engine supports ABFSS.
    parquet_abfss_path : str, optional
        Path to the parquet files in ABFSS. Must be the root directory containing a folder named after 
        each table in TABLE_REGISTRY.
    result_abfss_path : str, optional
        ABFSS path to the table where results will be saved. Must be specified if `save_results` is True.
    save_results : bool
        Whether to save the benchmark results. Results can also be accessed via the `self.results` 
        attribute after running the benchmark.

    Methods
    -------
    run(mode='power_test')
        Runs the benchmark in the specified mode.
        Supported modes are:
            - 'load': Sequentially executes loading the 24 tables.
            - 'query': Sequentially executes the 99 queries.
            - 'power_test': Executes the load test followed by the query test.
    _run_load_test()
        Loads the data for the benchmark.
    _run_query_test()
        Executes the queries for the benchmark.
    _run_power_test()
        Runs both the load and query tests.
    """
    TPC_BENCHMARK_VARIANT = 'DS'
    TABLE_REGISTRY = [
        'call_center', 'catalog_page', 'catalog_returns', 'catalog_sales',
        'customer', 'customer_address', 'customer_demographics', 'date_dim',
        'household_demographics', 'income_band', 'inventory', 'item',
        'promotion', 'reason', 'ship_mode', 'store', 'store_returns',
        'store_sales', 'time_dim', 'warehouse', 'web_page', 'web_returns',
        'web_sales', 'web_site'
    ]
    QUERY_REGISTRY = [
        'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8', 'q9', 'q10',
        'q11', 'q12', 'q13', 'q14a', 'q14b', 'q15', 'q16', 'q17', 'q18', 'q19', 'q20',
        'q21', 'q22', 'q23a', 'q23b', 'q24a', 'q24b', 'q25', 'q26', 'q27', 'q28', 'q29', 'q30',
        'q31', 'q32', 'q33', 'q34', 'q35', 'q36', 'q37', 'q38', 'q39a', 'q39b', 'q40',
        'q41', 'q42', 'q43', 'q44', 'q45', 'q46', 'q47', 'q48', 'q49', 'q50',
        'q51', 'q52', 'q53', 'q54', 'q55', 'q56', 'q57', 'q58', 'q59', 'q60',
        'q61', 'q62', 'q63', 'q64', 'q65', 'q66', 'q67', 'q68', 'q69', 'q70',
        'q71', 'q72', 'q73', 'q74', 'q75', 'q76', 'q77', 'q78', 'q79', 'q80',
        'q81', 'q82', 'q83', 'q84', 'q85', 'q86', 'q87', 'q88', 'q89', 'q90',
        'q91', 'q92', 'q93', 'q94', 'q95', 'q96', 'q97', 'q98', 'q99'
    ]
    DDL_FILE_NAME = 'ddl_v3.2.0.sql'